# -*- coding: utf-8 -*-
import warnings
warnings.simplefilter('always', DeprecationWarning)
warnings.warn('The mouse sub-library is deprecated and will be removed in future versions. Please use the standalone package `mouse`.', DeprecationWarning, stacklevel=2)

import time as _time

import platform as _platform
if _platform.system() == 'Windows':
    from. import _winmouse as _os_mouse
elif _platform.system() == 'Linux':
    from. import _nixmouse as _os_mouse
elif _platform.system() == 'Darwin':
    from. import _darwinmouse as _os_mouse
else:
    raise OSError("Unsupported platform '{}'".format(_platform.system()))

from ._mouse_event import ButtonEvent, MoveEvent, WheelEvent, LEFT, RIGHT, MIDDLE, X, X2, UP, DOWN, DOUBLE
from ._generic import GenericListener as _GenericListener

_pressed_events = set()
class _MouseListener(_GenericListener):
    def init(self):
        _os_mouse.init()
    def pre_process_event(self, event):
        if isinstance(event, ButtonEvent):
            if event.event_type in (UP, DOUBLE):
                _pressed_events.discard(event.button)
            else:
                _pressed_events.add(event.button)
        return True

    def listen(self):
        _os_mouse.listen(self.queue)

_listener = _MouseListener()

def is_pressed(button=LEFT):
    """ Returns True if the given button is currently pressed. """
    _listener.start_if_necessary()
    return button in _pressed_events

def press(button=LEFT):
    """ Presses the given button (but doesn't release). """
    _os_mouse.press(button)

def release(button=LEFT):
    """ Releases the given button. """
    _os_mouse.release(button)

def click(button=LEFT):
    """ Sends a click with the given button. """
    _os_mouse.press(button)
    _os_mouse.release(button)

def double_click(button=LEFT):
    """ Sends a double click with the given button. """
    click(button)
    click(button)

def right_click():
    """ Sends a right click with the given button. """
    click(RIGHT)

def wheel(delta=1):
    """ Scrolls the wheel `delta` clicks. Sign indicates direction. """
    _os_mouse.wheel(delta)

def move(x, y, absolute=True, duration=0):
    """
    Moves the mouse. If `absolute`, to position (x, y), otherwise move relative
    to the current position. If `duration` is non-zero, animates the movement.
    """
    x = int(x)
    y = int(y)

    # Requires an extra system call on Linux, but `move_relative` is measured
    # in millimeters so we would lose precision.
    position_x, position_y = get_position()

    if not absolute:
        x = position_x + x
        y = position_y + y

    if duration:
        start_x = position_x
        start_y = position_y
        dx = x - start_x
        dy = y - start_y

        if dx == 0 and dy == 0:
            _time.sleep(duration)
        else:
            # 120 movements per second.
            # Round and keep float to ensure float division in Python 2
            steps = max(1.0, float(int(duration * 120.0)))
            for i in range(int(steps)+1):
                move(start_x + dx*i/steps, start_y + dy*i/steps)
                _time.sleep(duration/steps)
    else:
        _os_mouse.move_to(x, y)

def drag(start_x, start_y, end_x, end_y, absolute=True, duration=0):
    """
    Holds the left mouse button, moving from start to end position, then
    releases. `absolute` and `duration` are parameters regarding the mouse
    movement.
    """
    if is_pressed():
        release()
    move(start_x, start_y, absolute, 0)
    press()
    move(end_x, end_y, absolute, duration)
    release()

def on_button(callback, args=(), buttons=(LEFT, MIDDLE, RIGHT, X, X2), types=(UP, DOWN, DOUBLE)):
    """ Invokes `callback` with `args` when the specified event happens. """
    if not isinstance(buttons, (tuple, list)):
        buttons = (buttons,)
    if not isinstance(types, (tuple, list)):
        types = (types,)

    def handler(event):
        if isinstance(event, ButtonEvent):
            if event.event_type in types and event.button in buttons:
                callback(*args)
    _listener.add_handler(handler)
    return handler

def on_click(callback, args=()):
    """ Invokes `callback` with `args` when the left button is clicked. """
    return on_button(callback, args, [LEFT], [UP])

def on_double_click(callback, args=()):
    """
    Invokes `callback` with `args` when the left button is double clicked.
    """
    return on_button(callback, args, [LEFT], [DOUBLE])

def on_right_click(callback, args=()):
    """ Invokes `callback` with `args` when the right button is clicked. """
    return on_button(callback, args, [RIGHT], [UP])

def on_middle_click(callback, args=()):
    """ Invokes `callback` with `args` when the middle button is clicked. """
    return on_button(callback, args, [MIDDLE], [UP])

def wait(button=LEFT, target_types=(UP, DOWN, DOUBLE)):
    """
    Blocks program execution until the given button performs an event.
    """
    from threading import Lock
    lock = Lock()
    lock.acquire()
    handler = on_button(lock.release, (), [button], target_types)
    lock.acquire()
    _listener.remove_handler(handler)

def get_position():
    """ Returns the (x, y) mouse position. """
    return _os_mouse.get_position()

def hook(callback):
    """
    Installs a global listener on all available mouses, invoking `callback`
    each time it is moved, a key status changes or the wheel is spun. A mouse
    event is passed as argument, with type either `mouse.ButtonEvent`,
    `mouse.WheelEvent` or `mouse.MoveEvent`.
    
    Returns the given callback for easier development.
    """
    _listener.add_handler(callback)
    return callback

def unhook(callback):
    """
    Removes a previously installed hook.
    """
    _listener.remove_handler(callback)

def unhook_all():
    """
    Removes all hooks registered by this application. Note this may include
    hooks installed by high level functions, such as `record`.
    """
    del _listener.handlers[:]

def record(button=RIGHT, target_types=(DOWN,)):
    """
    Records all mouse events until the user presses the given button.
    Then returns the list of events recorded. Pairs well with `play(events)`.

    Note: this is a blocking function.
    Note: for more details on the mouse hook and events see `hook`.
    """
    recorded = []
    hook(recorded.append)
    wait(button=button, target_types=target_types)
    unhook(recorded.append)
    return recorded

def play(events, speed_factor=1.0, include_clicks=True, include_moves=True, include_wheel=True):
    """
    Plays a sequence of recorded events, maintaining the relative time
    intervals. If speed_factor is <= 0 then the actions are replayed as fast
    as the OS allows. Pairs well with `record()`.

    The parameters `include_*` define if events of that type should be included
    in the replay or ignored.
    """
    last_time = None
    for event in events:
        if speed_factor > 0 and last_time is not None:
            _time.sleep((event.time - last_time) / speed_factor)
        last_time = event.time

        if isinstance(event, ButtonEvent) and include_clicks:
            if event.event_type == UP:
                _os_mouse.release(event.button)
            else:
                _os_mouse.press(event.button)
        elif isinstance(event, MoveEvent) and include_moves:
            _os_mouse.move_to(event.x, event.y)
        elif isinstance(event, WheelEvent) and include_wheel:
            _os_mouse.wheel(event.delta)

replay = play
hold = press

if __name__ == '__main__':
    print('Recording... Double click to stop and replay.')
    play(record())



# -*- coding: utf-8 -*-
import struct
import os
import atexit
from time import time as now
from threading import Thread
from glob import glob
try:
    from queue import Queue
except ImportError:
    from Queue import Queue

event_bin_format = 'llHHI'

# Taken from include/linux/input.h
# https://www.kernel.org/doc/Documentation/input/event-codes.txt
EV_SYN = 0x00
EV_KEY = 0x01
EV_REL = 0x02
EV_ABS = 0x03
EV_MSC = 0x04

def make_uinput():
    if not os.path.exists('/dev/uinput'):
        raise IOError('No uinput module found.')

    import fcntl, struct

    # Requires uinput driver, but it's usually available.
    uinput = open("/dev/uinput", 'wb')
    UI_SET_EVBIT = 0x40045564
    fcntl.ioctl(uinput, UI_SET_EVBIT, EV_KEY)

    UI_SET_KEYBIT = 0x40045565
    for i in range(256):
        fcntl.ioctl(uinput, UI_SET_KEYBIT, i)

    BUS_USB = 0x03
    uinput_user_dev = "80sHHHHi64i64i64i64i"
    axis = [0] * 64 * 4
    uinput.write(struct.pack(uinput_user_dev, b"Virtual Keyboard", BUS_USB, 1, 1, 1, 0, *axis))
    uinput.flush() # Without this you may get Errno 22: Invalid argument.

    UI_DEV_CREATE = 0x5501
    fcntl.ioctl(uinput, UI_DEV_CREATE)
    UI_DEV_DESTROY = 0x5502
    #fcntl.ioctl(uinput, UI_DEV_DESTROY)

    return uinput

class EventDevice(object):
    def __init__(self, path):
        self.path = path
        self._input_file = None
        self._output_file = None

    @property
    def input_file(self):
        if self._input_file is None:
            try:
                self._input_file = open(self.path, 'rb')
            except IOError as e:
                if e.strerror == 'Permission denied':
                    print("# ERROR: Failed to read device '{}'. You must be in the 'input' group to access global events. Use 'sudo usermod -a -G input USERNAME' to add user to the required group.".format(self.path))
                    exit()

            def try_close():
                try:
                    self._input_file.close
                except:
                    pass
            atexit.register(try_close)
        return self._input_file

    @property
    def output_file(self):
        if self._output_file is None:
            self._output_file = open(self.path, 'wb')
            atexit.register(self._output_file.close)
        return self._output_file

    def read_event(self):
        data = self.input_file.read(struct.calcsize(event_bin_format))
        seconds, microseconds, type, code, value = struct.unpack(event_bin_format, data)
        return seconds + microseconds / 1e6, type, code, value, self.path

    def write_event(self, type, code, value):
        integer, fraction = divmod(now(), 1)
        seconds = int(integer)
        microseconds = int(fraction * 1e6)
        data_event = struct.pack(event_bin_format, seconds, microseconds, type, code, value)

        # Send a sync event to ensure other programs update.
        sync_event = struct.pack(event_bin_format, seconds, microseconds, EV_SYN, 0, 0)

        self.output_file.write(data_event + sync_event)
        self.output_file.flush()

class AggregatedEventDevice(object):
    def __init__(self, devices, output=None):
        self.event_queue = Queue()
        self.devices = devices
        self.output = output or self.devices[0]
        def start_reading(device):
            while True:
                self.event_queue.put(device.read_event())
        for device in self.devices:
            thread = Thread(target=start_reading, args=[device])
            thread.daemon = True
            thread.start()

    def read_event(self):
        return self.event_queue.get(block=True)

    def write_event(self, type, code, value):
        self.output.write_event(type, code, value)

import re
from collections import namedtuple
DeviceDescription = namedtuple('DeviceDescription', 'event_file is_mouse is_keyboard')
device_pattern = r"""N: Name="([^"]+?)".+?H: Handlers=([^\n]+)"""
def list_devices_from_proc(type_name):
    try:
        with open('/proc/bus/input/devices') as f:
            description = f.read()
    except FileNotFoundError:
        return

    devices = {}
    for name, handlers in re.findall(device_pattern, description, re.DOTALL):
        path = '/dev/input/event' + re.search(r'event(\d+)', handlers).group(1)
        if type_name in handlers:
            yield EventDevice(path)

def list_devices_from_by_id(name_suffix, by_id=True):
    for path in glob('/dev/input/{}/*-event-{}'.format('by-id' if by_id else 'by-path', name_suffix)):
        yield EventDevice(path)

def aggregate_devices(type_name):
    # Some systems have multiple keyboards with different range of allowed keys
    # on each one, like a notebook with a "keyboard" device exclusive for the
    # power button. Instead of figuring out which keyboard allows which key to
    # send events, we create a fake device and send all events through there.
    try:
        uinput = make_uinput()
        fake_device = EventDevice('uinput Fake Device')
        fake_device._input_file = uinput
        fake_device._output_file = uinput
    except IOError as e:
        import warnings
        warnings.warn('Failed to create a device file using `uinput` module. Sending of events may be limited or unavailable depending on plugged-in devices.', stacklevel=2)
        fake_device = None

    # We don't aggregate devices from different sources to avoid
    # duplicates.

    devices_from_proc = list(list_devices_from_proc(type_name))
    if devices_from_proc:
        return AggregatedEventDevice(devices_from_proc, output=fake_device)

    # breaks on mouse for virtualbox
    # was getting /dev/input/by-id/usb-VirtualBox_USB_Tablet-event-mouse
    devices_from_by_id = list(list_devices_from_by_id(type_name)) or list(list_devices_from_by_id(type_name, by_id=False))
    if devices_from_by_id:
        return AggregatedEventDevice(devices_from_by_id, output=fake_device)

    # If no keyboards were found we can only use the fake device to send keys.
    assert fake_device
    return fake_device



# -*- coding: utf-8 -*-

from time import time as now
import json
from ._canonical_names import canonical_names, normalize_name

try:
    basestring
except NameError:
    basestring = str

KEY_DOWN = 'down'
KEY_UP = 'up'

class KeyboardEvent(object):
    event_type = None
    scan_code = None
    name = None
    time = None
    device = None
    modifiers = None
    is_keypad = None

    def __init__(self, event_type, scan_code, name=None, time=None, device=None, modifiers=None, is_keypad=None):
        self.event_type = event_type
        self.scan_code = scan_code
        self.time = now() if time is None else time
        self.device = device
        self.is_keypad = is_keypad
        self.modifiers = modifiers
        if name:
            self.name = normalize_name(name)

    def to_json(self, ensure_ascii=False):
        attrs = dict(
            (attr, getattr(self, attr)) for attr in ['event_type', 'scan_code', 'name', 'time', 'device', 'is_keypad', 'modifiers']
            if not attr.startswith('_')
        )
        return json.dumps(attrs, ensure_ascii=ensure_ascii)

    def __repr__(self):
        return 'KeyboardEvent({} {})'.format(self.name or 'Unknown {}'.format(self.scan_code), self.event_type)

    def __eq__(self, other):
        return (
            isinstance(other, KeyboardEvent)
            and self.event_type == other.event_type
            and (
                not self.scan_code or not other.scan_code or self.scan_code == other.scan_code
            ) and (
                not self.name or not other.name or self.name == other.name
            )
        )



# -*- coding: utf-8 -*-
from threading import Thread, Lock
import traceback
import functools

try:
    from queue import Queue
except ImportError:
    from Queue import Queue

class GenericListener(object):
    lock = Lock()

    def __init__(self):
        self.handlers = []
        self.listening = False
        self.queue = Queue()

    def invoke_handlers(self, event):
        for handler in self.handlers:
            try:
                if handler(event):
                    # Stop processing this hotkey.
                    return 1
            except Exception as e:
                traceback.print_exc()

    def start_if_necessary(self):
        """
        Starts the listening thread if it wasn't already.
        """
        self.lock.acquire()
        try:
            if not self.listening:
                self.init()

                self.listening = True
                self.listening_thread = Thread(target=self.listen)
                self.listening_thread.daemon = True
                self.listening_thread.start()

                self.processing_thread = Thread(target=self.process)
                self.processing_thread.daemon = True
                self.processing_thread.start()
        finally:
            self.lock.release()

    def pre_process_event(self, event):
        raise NotImplementedError('This method should be implemented in the child class.')

    def process(self):
        """
        Loops over the underlying queue of events and processes them in order.
        """
        assert self.queue is not None
        while True:
            event = self.queue.get()
            if self.pre_process_event(event):
                self.invoke_handlers(event)
            self.queue.task_done()
            
    def add_handler(self, handler):
        """
        Adds a function to receive each event captured, starting the capturing
        process if necessary.
        """
        self.start_if_necessary()
        self.handlers.append(handler)

    def remove_handler(self, handler):
        """ Removes a previously added event handler. """
        while handler in self.handlers:
            self.handlers.remove(handler)



# -*- coding: utf-8 -*-
import struct
import traceback
from time import time as now
from collections import namedtuple
from ._keyboard_event import KeyboardEvent, KEY_DOWN, KEY_UP
from ._canonical_names import all_modifiers, normalize_name
from ._nixcommon import EV_KEY, aggregate_devices

# TODO: start by reading current keyboard state, as to not missing any already pressed keys.
# See: http://stackoverflow.com/questions/3649874/how-to-get-keyboard-state-in-linux

def cleanup_key(name):
    """ Formats a dumpkeys format to our standard. """
    name = name.lstrip('+')
    is_keypad = name.startswith('KP_')
    for mod in ('Meta_', 'Control_', 'dead_', 'KP_'):
        if name.startswith(mod):
            name = name[len(mod):]

    # Dumpkeys is weird like that.
    if name == 'Remove':
        name = 'Delete'
    elif name == 'Delete':
        name = 'Backspace'

    if name.endswith('_r'):
        name = 'right ' + name[:-2]
    if name.endswith('_l'):
        name = 'left ' + name[:-2]


    return normalize_name(name), is_keypad

def cleanup_modifier(modifier):
    modifier = normalize_name(modifier)
    if modifier in all_modifiers:
        return modifier
    if modifier[:-1] in all_modifiers:
        return modifier[:-1]
    raise ValueError('Unknown modifier {}'.format(modifier))

"""
Use `dumpkeys --keys-only` to list all scan codes and their names. We
then parse the output and built a table. For each scan code and modifiers we
have a list of names and vice-versa.
"""
from subprocess import check_output, CalledProcessError, PIPE
from collections import defaultdict
import re

to_name = defaultdict(list)
from_name = defaultdict(list)
keypad_scan_codes = set()

def register_key(key_and_modifiers, name):
    if name not in to_name[key_and_modifiers]:
        to_name[key_and_modifiers].append(name)
    if key_and_modifiers not in from_name[name]:
        from_name[name].append(key_and_modifiers)

def build_tables():
    if to_name and from_name: return

    modifiers_bits = {
        'shift': 1,
        'alt gr': 2,
        'ctrl': 4,
        'alt': 8,
    }
    keycode_template = r'^keycode\s+(\d+)\s+=(.*?)$'
    try:
        dump = check_output(['dumpkeys', '--keys-only'], universal_newlines=True)
    except CalledProcessError as e:
        if e.returncode == 1:
            raise ValueError('Failed to run dumpkeys to get key names. Check if your user is part of the "tty" group, and if not, add it with "sudo usermod -a -G tty USER".')
        else:
            raise


    for str_scan_code, str_names in re.findall(keycode_template, dump, re.MULTILINE):
        scan_code = int(str_scan_code)
        for i, str_name in enumerate(str_names.strip().split()):
            modifiers = tuple(sorted(modifier for modifier, bit in modifiers_bits.items() if i & bit))
            name, is_keypad = cleanup_key(str_name)
            register_key((scan_code, modifiers), name)
            if is_keypad:
                keypad_scan_codes.add(scan_code)
                register_key((scan_code, modifiers), 'keypad ' + name)

    # dumpkeys consistently misreports the Windows key, sometimes
    # skipping it completely or reporting as 'alt. 125 = left win,
    # 126 = right win.
    if (125, ()) not in to_name or to_name[(125, ())] == ['alt']:
        to_name[(125, ())].clear()
        if (125, ()) in from_name['alt']:
            from_name['alt'].remove((125, ()))
        register_key((125, ()), 'windows')
    if (126, ()) not in to_name or to_name[(126, ())] == ['alt']:
        to_name[(126, ())].clear()
        if (126, ()) in from_name['alt']:
            from_name['alt'].remove((126, ()))
        register_key((126, ()), 'windows')

    # The menu key is usually skipped altogether, so we also add it manually.
    if (127, ()) not in to_name:
        register_key((127, ()), 'menu')

    synonyms_template = r'^(\S+)\s+for (.+)$'
    dump = check_output(['dumpkeys', '--long-info'], universal_newlines=True)
    for synonym_str, original_str in re.findall(synonyms_template, dump, re.MULTILINE):
        synonym, _ = cleanup_key(synonym_str)
        original, _ = cleanup_key(original_str)
        if synonym != original:
            from_name[original].extend(from_name[synonym])
            from_name[synonym].extend(from_name[original])

device = None
def build_device():
    global device
    if device: return
    device = aggregate_devices('kbd')

def init():
    build_device()
    build_tables()

pressed_modifiers = set()

def listen(callback):
    build_device()
    build_tables()

    while True:
        time, type, code, value, device_id = device.read_event()
        if type != EV_KEY:
            continue

        scan_code = code
        event_type = KEY_DOWN if value else KEY_UP # 0 = UP, 1 = DOWN, 2 = HOLD

        pressed_modifiers_tuple = tuple(sorted(pressed_modifiers))
        names = to_name[(scan_code, pressed_modifiers_tuple)] or to_name[(scan_code, ())] or ['unknown']
        name = names[0]
            
        if name in all_modifiers:
            if event_type == KEY_DOWN:
                pressed_modifiers.add(name)
            else:
                pressed_modifiers.discard(name)

        is_keypad = scan_code in keypad_scan_codes
        callback(KeyboardEvent(event_type=event_type, scan_code=scan_code, name=name, time=time, device=device_id, is_keypad=is_keypad, modifiers=pressed_modifiers_tuple))

def write_event(scan_code, is_down):
    build_device()
    device.write_event(EV_KEY, scan_code, int(is_down))

def map_name(name):
    build_tables()
    for entry in from_name[name]:
        yield entry

    parts = name.split(' ', 1)
    if len(parts) > 1 and parts[0] in ('left', 'right'):
        for entry in from_name[parts[1]]:
            yield entry

def press(scan_code):
    write_event(scan_code, True)

def release(scan_code):
    write_event(scan_code, False)

def type_unicode(character):
    codepoint = ord(character)
    hexadecimal = hex(codepoint)[len('0x'):]

    for key in ['ctrl', 'shift', 'u']:
        scan_code, _ = next(map_name(key))
        press(scan_code)

    for key in hexadecimal:
        scan_code, _ = next(map_name(key))
        press(scan_code)
        release(scan_code)

    for key in ['ctrl', 'shift', 'u']:
        scan_code, _ = next(map_name(key))
        release(scan_code)

if __name__ == '__main__':
    def p(e):
        print(e)
    listen(p)



"""
Usage instructions:

- If you are installing: `python setup.py install`
- If you are developing: `python setup.py sdist --format=zip bdist_wheel --universal bdist_wininst && twine check dist/*`
"""
import keyboard

from setuptools import setup
setup(
    name='keyboard-darwin-fix',
    version=keyboard.version,
    author='kkristof200',
    author_email='',
    packages=['keyboard'],
    url='https://github.com/kkristof200/keyboard-darwin-fix',
    license='MIT',
    description='Hook and simulate keyboard events on Windows and Linux',
    keywords = 'keyboard hook simulate hotkey',

    # Wheel creation breaks with Windows newlines.
    # https://github.com/pypa/setuptools/issues/1126
    long_description=keyboard.__doc__.replace('\r\n', '\n'),
    long_description_content_type='text/markdown',

    install_requires=["pyobjc; sys_platform=='darwin'"], # OSX-specific dependency
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
)



# -*- coding: utf-8 -*-
from collections import namedtuple

LEFT = 'left'
RIGHT = 'right'
MIDDLE = 'middle'
WHEEL = 'wheel'
X = 'x'
X2 = 'x2'

UP = 'up'
DOWN = 'down'
DOUBLE = 'double'
VERTICAL = 'vertical'
HORIZONTAL = 'horizontal'


ButtonEvent = namedtuple('ButtonEvent', ['event_type', 'button', 'time'])
WheelEvent = namedtuple('WheelEvent', ['delta', 'time'])
MoveEvent = namedtuple('MoveEvent', ['x', 'y', 'time'])



# -*- coding: utf-8 -*-
"""
Side effects are avoided using two techniques:

- Low level OS requests (keyboard._os_keyboard) are mocked out by rewriting
the functions at that namespace. This includes a list of dummy keys.
- Events are pumped manually by the main test class, and accepted events
are tested against expected values.

Fake user events are appended to `input_events`, passed through
keyboard,_listener.direct_callback, then, if accepted, appended to
`output_events`. Fake OS events (keyboard.press) are processed
and added to `output_events` immediately, mimicking real functionality.
"""
from __future__ import print_function

import unittest
import time

import keyboard
from ._keyboard_event import KeyboardEvent, KEY_DOWN, KEY_UP

dummy_keys = {
    'space': [(0, [])],

    'a': [(1, [])],
    'b': [(2, [])],
    'c': [(3, [])],
    'A': [(1, ['shift']), (-1, [])],
    'B': [(2, ['shift']), (-2, [])],
    'C': [(3, ['shift']), (-3, [])],

    'alt': [(4, [])],
    'left alt': [(4, [])],

    'left shift': [(5, [])],
    'right shift': [(6, [])],

    'left ctrl': [(7, [])],

    'backspace': [(8, [])],
    'caps lock': [(9, [])],

    '+': [(10, [])],
    ',': [(11, [])],
    '_': [(12, [])],

    'none': [],
    'duplicated': [(20, []), (20, [])],
}

def make_event(event_type, name, scan_code=None, time=0):
    return KeyboardEvent(event_type=event_type, scan_code=scan_code or dummy_keys[name][0][0], name=name, time=time)

# Used when manually pumping events.
input_events = []
output_events = []

def send_instant_event(event):
    if keyboard._listener.direct_callback(event):
        output_events.append(event)

# Mock out side effects.
keyboard._os_keyboard.init = lambda: None
keyboard._os_keyboard.listen = lambda callback: None
keyboard._os_keyboard.map_name = dummy_keys.__getitem__
keyboard._os_keyboard.press = lambda scan_code: send_instant_event(make_event(KEY_DOWN, None, scan_code))
keyboard._os_keyboard.release = lambda scan_code: send_instant_event(make_event(KEY_UP, None, scan_code))
keyboard._os_keyboard.type_unicode = lambda char: output_events.append(KeyboardEvent(event_type=KEY_DOWN, scan_code=999, name=char))

# Shortcuts for defining test inputs and expected outputs.
# Usage: d_shift + d_a + u_a + u_shift
d_a = [make_event(KEY_DOWN, 'a')]
u_a = [make_event(KEY_UP, 'a')]
du_a = d_a+u_a
d_b = [make_event(KEY_DOWN, 'b')]
u_b = [make_event(KEY_UP, 'b')]
du_b = d_b+u_b
d_c = [make_event(KEY_DOWN, 'c')]
u_c = [make_event(KEY_UP, 'c')]
du_c = d_c+u_c
d_ctrl = [make_event(KEY_DOWN, 'left ctrl')]
u_ctrl = [make_event(KEY_UP, 'left ctrl')]
du_ctrl = d_ctrl+u_ctrl
d_shift = [make_event(KEY_DOWN, 'left shift')]
u_shift = [make_event(KEY_UP, 'left shift')]
du_shift = d_shift+u_shift
d_alt = [make_event(KEY_DOWN, 'alt')]
u_alt = [make_event(KEY_UP, 'alt')]
du_alt = d_alt+u_alt
du_backspace = [make_event(KEY_DOWN, 'backspace'), make_event(KEY_UP, 'backspace')]
du_capslock = [make_event(KEY_DOWN, 'caps lock'), make_event(KEY_UP, 'caps lock')]
d_space = [make_event(KEY_DOWN, 'space')]
u_space = [make_event(KEY_UP, 'space')]
du_space = [make_event(KEY_DOWN, 'space'), make_event(KEY_UP, 'space')]

trigger = lambda e=None: keyboard.press(999)
triggered_event = [KeyboardEvent(KEY_DOWN, scan_code=999)]

class TestKeyboard(unittest.TestCase):
    def tearDown(self):
        keyboard.unhook_all()
        #self.assertEquals(keyboard._hooks, {})
        #self.assertEquals(keyboard._hotkeys, {})

    def setUp(self):
        #keyboard._hooks.clear()
        #keyboard._hotkeys.clear()
        del input_events[:]
        del output_events[:]
        keyboard._recording = None
        keyboard._pressed_events.clear()
        keyboard._physically_pressed_keys.clear()
        keyboard._logically_pressed_keys.clear()
        keyboard._hotkeys.clear()
        keyboard._listener.init()
        keyboard._word_listeners = {} 

    def do(self, manual_events, expected=None):
        input_events.extend(manual_events)
        while input_events:
            event = input_events.pop(0)
            if keyboard._listener.direct_callback(event):
                output_events.append(event)
        if expected is not None:
            to_names = lambda es: '+'.join(('d' if e.event_type == KEY_DOWN else 'u') + '_' + str(e.scan_code) for e in es)
            self.assertEqual(to_names(output_events), to_names(expected))
        del output_events[:]

        keyboard._listener.queue.join()

    def test_event_json(self):
        event = make_event(KEY_DOWN, u'á \'"', 999)
        import json
        self.assertEqual(event, KeyboardEvent(**json.loads(event.to_json())))

    def test_is_modifier_name(self):
        for name in keyboard.all_modifiers:
            self.assertTrue(keyboard.is_modifier(name))
    def test_is_modifier_scan_code(self):
        for i in range(10):
            self.assertEqual(keyboard.is_modifier(i), i in [4, 5, 6, 7])

    def test_key_to_scan_codes_brute(self):
        for name, entries in dummy_keys.items():
            if name in ['none', 'duplicated']: continue
            expected = tuple(scan_code for scan_code, modifiers in entries)
            self.assertEqual(keyboard.key_to_scan_codes(name), expected)
    def test_key_to_scan_code_from_scan_code(self):
        for i in range(10):
            self.assertEqual(keyboard.key_to_scan_codes(i), (i,))
    def test_key_to_scan_code_from_letter(self):
        self.assertEqual(keyboard.key_to_scan_codes('a'), (1,))
        self.assertEqual(keyboard.key_to_scan_codes('A'), (1,-1))
    def test_key_to_scan_code_from_normalized(self):
        self.assertEqual(keyboard.key_to_scan_codes('shift'), (5,6))
        self.assertEqual(keyboard.key_to_scan_codes('SHIFT'), (5,6))
        self.assertEqual(keyboard.key_to_scan_codes('ctrl'), keyboard.key_to_scan_codes('CONTROL'))
    def test_key_to_scan_code_from_sided_modifier(self):
        self.assertEqual(keyboard.key_to_scan_codes('left shift'), (5,))
        self.assertEqual(keyboard.key_to_scan_codes('right shift'), (6,))
    def test_key_to_scan_code_underscores(self):
        self.assertEqual(keyboard.key_to_scan_codes('_'), (12,))
        self.assertEqual(keyboard.key_to_scan_codes('right_shift'), (6,))
    def test_key_to_scan_code_error_none(self):
        with self.assertRaises(ValueError):
            keyboard.key_to_scan_codes(None)
    def test_key_to_scan_code_error_empty(self):
        with self.assertRaises(ValueError):
            keyboard.key_to_scan_codes('')
    def test_key_to_scan_code_error_other(self):
        with self.assertRaises(ValueError):
            keyboard.key_to_scan_codes({})
    def test_key_to_scan_code_list(self):
        self.assertEqual(keyboard.key_to_scan_codes([10, 5, 'a']), (10, 5, 1))
    def test_key_to_scan_code_empty(self):
        with self.assertRaises(ValueError):
            keyboard.key_to_scan_codes('none')
    def test_key_to_scan_code_duplicated(self):
        self.assertEqual(keyboard.key_to_scan_codes('duplicated'), (20,))

    def test_parse_hotkey_simple(self):
        self.assertEqual(keyboard.parse_hotkey('a'), (((1,),),))
        self.assertEqual(keyboard.parse_hotkey('A'), (((1,-1),),))
    def test_parse_hotkey_separators(self):
        self.assertEqual(keyboard.parse_hotkey('+'), keyboard.parse_hotkey('plus'))
        self.assertEqual(keyboard.parse_hotkey(','), keyboard.parse_hotkey('comma'))
    def test_parse_hotkey_keys(self):
        self.assertEqual(keyboard.parse_hotkey('left shift + a'), (((5,), (1,),),))
        self.assertEqual(keyboard.parse_hotkey('left shift+a'), (((5,), (1,),),))
    def test_parse_hotkey_simple_steps(self):
        self.assertEqual(keyboard.parse_hotkey('a,b'), (((1,),),((2,),)))
        self.assertEqual(keyboard.parse_hotkey('a, b'), (((1,),),((2,),)))
    def test_parse_hotkey_steps(self):
        self.assertEqual(keyboard.parse_hotkey('a+b, b+c'), (((1,),(2,)),((2,),(3,))))
    def test_parse_hotkey_example(self):
        alt_codes = keyboard.key_to_scan_codes('alt')
        shift_codes = keyboard.key_to_scan_codes('shift')
        a_codes = keyboard.key_to_scan_codes('a')
        b_codes = keyboard.key_to_scan_codes('b')
        c_codes = keyboard.key_to_scan_codes('c')
        self.assertEqual(keyboard.parse_hotkey("alt+shift+a, alt+b, c"), ((alt_codes, shift_codes, a_codes), (alt_codes, b_codes), (c_codes,)))
    def test_parse_hotkey_list_scan_codes(self):
        self.assertEqual(keyboard.parse_hotkey([1, 2, 3]), (((1,), (2,), (3,)),))
    def test_parse_hotkey_deep_list_scan_codes(self):
        result = keyboard.parse_hotkey('a')
        self.assertEqual(keyboard.parse_hotkey(result), (((1,),),))
    def test_parse_hotkey_list_names(self):
        self.assertEqual(keyboard.parse_hotkey(['a', 'b', 'c']), (((1,), (2,), (3,)),))

    def test_is_pressed_none(self):
        self.assertFalse(keyboard.is_pressed('a'))
    def test_is_pressed_true(self):
        self.do(d_a)
        self.assertTrue(keyboard.is_pressed('a'))
    def test_is_pressed_true_scan_code_true(self):
        self.do(d_a)
        self.assertTrue(keyboard.is_pressed(1))
    def test_is_pressed_true_scan_code_false(self):
        self.do(d_a)
        self.assertFalse(keyboard.is_pressed(2))
    def test_is_pressed_true_scan_code_invalid(self):
        self.do(d_a)
        self.assertFalse(keyboard.is_pressed(-1))
    def test_is_pressed_false(self):
        self.do(d_a+u_a+d_b)
        self.assertFalse(keyboard.is_pressed('a'))
        self.assertTrue(keyboard.is_pressed('b'))
    def test_is_pressed_hotkey_true(self):
        self.do(d_shift+d_a)
        self.assertTrue(keyboard.is_pressed('shift+a'))
    def test_is_pressed_hotkey_false(self):
        self.do(d_shift+d_a+u_a)
        self.assertFalse(keyboard.is_pressed('shift+a'))
    def test_is_pressed_multi_step_fail(self):
        self.do(u_a+d_a)
        with self.assertRaises(ValueError):
            keyboard.is_pressed('a, b')

    def test_send_single_press_release(self):
        keyboard.send('a', do_press=True, do_release=True)
        self.do([], d_a+u_a)
    def test_send_single_press(self):
        keyboard.send('a', do_press=True, do_release=False)
        self.do([], d_a)
    def test_send_single_release(self):
        keyboard.send('a', do_press=False, do_release=True)
        self.do([], u_a)
    def test_send_single_none(self):
        keyboard.send('a', do_press=False, do_release=False)
        self.do([], [])
    def test_press(self):
        keyboard.press('a')
        self.do([], d_a)
    def test_release(self):
        keyboard.release('a')
        self.do([], u_a)
    def test_press_and_release(self):
        keyboard.press_and_release('a')
        self.do([], d_a+u_a)

    def test_send_modifier_press_release(self):
        keyboard.send('ctrl+a', do_press=True, do_release=True)
        self.do([], d_ctrl+d_a+u_a+u_ctrl)
    def test_send_modifiers_release(self):
        keyboard.send('ctrl+shift+a', do_press=False, do_release=True)
        self.do([], u_a+u_shift+u_ctrl)

    def test_call_later(self):
        triggered = []
        def fn(arg1, arg2):
            assert arg1 == 1 and arg2 == 2
            triggered.append(True)
        keyboard.call_later(fn, (1, 2), 0.01)
        self.assertFalse(triggered)
        time.sleep(0.05)
        self.assertTrue(triggered)

    def test_hook_nonblocking(self):
        self.i = 0
        def count(e):
            self.assertEqual(e.name, 'a')
            self.i += 1
        hook = keyboard.hook(count, suppress=False)
        self.do(d_a+u_a, d_a+u_a)
        self.assertEqual(self.i, 2)
        keyboard.unhook(hook)
        self.do(d_a+u_a, d_a+u_a)
        self.assertEqual(self.i, 2)
        keyboard.hook(count, suppress=False)
        self.do(d_a+u_a, d_a+u_a)
        self.assertEqual(self.i, 4)
        keyboard.unhook_all()
        self.do(d_a+u_a, d_a+u_a)
        self.assertEqual(self.i, 4)
    def test_hook_blocking(self):
        self.i = 0
        def count(e):
            self.assertIn(e.name, ['a', 'b'])
            self.i += 1
            return e.name == 'b'
        hook = keyboard.hook(count, suppress=True)
        self.do(d_a+d_b, d_b)
        self.assertEqual(self.i, 2)
        keyboard.unhook(hook)
        self.do(d_a+d_b, d_a+d_b)
        self.assertEqual(self.i, 2)
        keyboard.hook(count, suppress=True)
        self.do(d_a+d_b, d_b)
        self.assertEqual(self.i, 4)
        keyboard.unhook_all()
        self.do(d_a+d_b, d_a+d_b)
        self.assertEqual(self.i, 4)
    def test_on_press_nonblocking(self):
        keyboard.on_press(lambda e: self.assertEqual(e.name, 'a') and self.assertEqual(e.event_type, KEY_DOWN))
        self.do(d_a+u_a)
    def test_on_press_blocking(self):
        keyboard.on_press(lambda e: e.scan_code == 1, suppress=True)
        self.do([make_event(KEY_DOWN, 'A', -1)] + d_a, d_a)
    def test_on_release(self):
        keyboard.on_release(lambda e: self.assertEqual(e.name, 'a') and self.assertEqual(e.event_type, KEY_UP))
        self.do(d_a+u_a)

    def test_hook_key_invalid(self):
        with self.assertRaises(ValueError):
            keyboard.hook_key('invalid', lambda e: None)
    def test_hook_key_nonblocking(self):
        self.i = 0
        def count(event):
            self.i += 1
        hook = keyboard.hook_key('A', count)
        self.do(d_a)
        self.assertEqual(self.i, 1)
        self.do(u_a+d_b)
        self.assertEqual(self.i, 2)
        self.do([make_event(KEY_DOWN, 'A', -1)])
        self.assertEqual(self.i, 3)
        keyboard.unhook_key(hook)
        self.do(d_a)
        self.assertEqual(self.i, 3)
    def test_hook_key_blocking(self):
        self.i = 0
        def count(event):
            self.i += 1
            return event.scan_code == 1
        hook = keyboard.hook_key('A', count, suppress=True)
        self.do(d_a, d_a)
        self.assertEqual(self.i, 1)
        self.do(u_a+d_b, u_a+d_b)
        self.assertEqual(self.i, 2)
        self.do([make_event(KEY_DOWN, 'A', -1)], [])
        self.assertEqual(self.i, 3)
        keyboard.unhook_key(hook)
        self.do([make_event(KEY_DOWN, 'A', -1)], [make_event(KEY_DOWN, 'A', -1)])
        self.assertEqual(self.i, 3)
    def test_on_press_key_nonblocking(self):
        keyboard.on_press_key('A', lambda e: self.assertEqual(e.name, 'a') and self.assertEqual(e.event_type, KEY_DOWN))
        self.do(d_a+u_a+d_b+u_b)
    def test_on_press_key_blocking(self):
        keyboard.on_press_key('A', lambda e: e.scan_code == 1, suppress=True)
        self.do([make_event(KEY_DOWN, 'A', -1)] + d_a, d_a)
    def test_on_release_key(self):
        keyboard.on_release_key('a', lambda e: self.assertEqual(e.name, 'a') and self.assertEqual(e.event_type, KEY_UP))
        self.do(d_a+u_a)

    def test_block_key(self):
        blocked = keyboard.block_key('a')
        self.do(d_a+d_b, d_b)
        self.do([make_event(KEY_DOWN, 'A', -1)], [make_event(KEY_DOWN, 'A', -1)])
        keyboard.unblock_key(blocked)
        self.do(d_a+d_b, d_a+d_b)
    def test_block_key_ambiguous(self):
        keyboard.block_key('A')
        self.do(d_a+d_b, d_b)
        self.do([make_event(KEY_DOWN, 'A', -1)], [])

    def test_remap_key_simple(self):
        mapped = keyboard.remap_key('a', 'b')
        self.do(d_a+d_c+u_a, d_b+d_c+u_b)
        keyboard.unremap_key(mapped)
        self.do(d_a+d_c+u_a, d_a+d_c+u_a)
    def test_remap_key_ambiguous(self):
        keyboard.remap_key('A', 'b')
        self.do(d_a+d_b, d_b+d_b)
        self.do([make_event(KEY_DOWN, 'A', -1)], d_b)
    def test_remap_key_multiple(self):
        mapped = keyboard.remap_key('a', 'shift+b')
        self.do(d_a+d_c+u_a, d_shift+d_b+d_c+u_b+u_shift)
        keyboard.unremap_key(mapped)
        self.do(d_a+d_c+u_a, d_a+d_c+u_a)

    def test_stash_state(self):
        self.do(d_a+d_shift)
        self.assertEqual(sorted(keyboard.stash_state()), [1, 5])
        self.do([], u_a+u_shift)
    def test_restore_state(self):
        self.do(d_b)
        keyboard.restore_state([1, 5])
        self.do([], u_b+d_a+d_shift)
    def test_restore_modifieres(self):
        self.do(d_b)
        keyboard.restore_modifiers([1, 5])
        self.do([], u_b+d_shift)

    def test_write_simple(self):
        keyboard.write('a', exact=False)
        self.do([], d_a+u_a)
    def test_write_multiple(self):
        keyboard.write('ab', exact=False)
        self.do([], d_a+u_a+d_b+u_b)
    def test_write_modifiers(self):
        keyboard.write('Ab', exact=False)
        self.do([], d_shift+d_a+u_a+u_shift+d_b+u_b)
    # restore_state_after has been removed after the introduction of `restore_modifiers`.
    #def test_write_stash_not_restore(self):
    #    self.do(d_shift)
    #    keyboard.write('a', restore_state_after=False, exact=False)
    #    self.do([], u_shift+d_a+u_a)
    def test_write_stash_restore(self):
        self.do(d_shift)
        keyboard.write('a', exact=False)
        self.do([], u_shift+d_a+u_a+d_shift)
    def test_write_multiple(self):
        last_time = time.time()
        keyboard.write('ab', delay=0.01, exact=False)
        self.do([], d_a+u_a+d_b+u_b)
        self.assertGreater(time.time() - last_time, 0.015)
    def test_write_unicode_explicit(self):
        keyboard.write('ab', exact=True)
        self.do([], [KeyboardEvent(event_type=KEY_DOWN, scan_code=999, name='a'), KeyboardEvent(event_type=KEY_DOWN, scan_code=999, name='b')])
    def test_write_unicode_fallback(self):
        keyboard.write(u'áb', exact=False)
        self.do([], [KeyboardEvent(event_type=KEY_DOWN, scan_code=999, name=u'á')]+d_b+u_b)

    def test_start_stop_recording(self):
        keyboard.start_recording()
        self.do(d_a+u_a)
        self.assertEqual(keyboard.stop_recording(), d_a+u_a)
    def test_stop_recording_error(self):
        with self.assertRaises(ValueError):
            keyboard.stop_recording()

    def test_record(self):
        queue = keyboard._queue.Queue()
        def process():
            queue.put(keyboard.record('space', suppress=True))
        from threading import Thread
        t = Thread(target=process)
        t.daemon = True
        t.start()
        # 0.01s sleep failed once already. Better solutions?
        time.sleep(0.01)
        self.do(du_a+du_b+du_space, du_a+du_b)
        self.assertEqual(queue.get(timeout=0.5), du_a+du_b+du_space)

    def test_play_nodelay(self):
        keyboard.play(d_a+u_a, 0)
        self.do([], d_a+u_a)
    def test_play_stash(self):
        self.do(d_ctrl)
        keyboard.play(d_a+u_a, 0)
        self.do([], u_ctrl+d_a+u_a+d_ctrl)
    def test_play_delay(self):
        last_time = time.time()
        events = [make_event(KEY_DOWN, 'a', 1, 100), make_event(KEY_UP, 'a', 1, 100.01)]
        keyboard.play(events, 1)
        self.do([], d_a+u_a)
        self.assertGreater(time.time() - last_time, 0.005)

    def test_get_typed_strings_simple(self):
        events = du_a+du_b+du_backspace+d_shift+du_a+u_shift+du_space+du_ctrl+du_a
        self.assertEqual(list(keyboard.get_typed_strings(events)), ['aA ', 'a'])
    def test_get_typed_strings_backspace(self):
        events = du_a+du_b+du_backspace
        self.assertEqual(list(keyboard.get_typed_strings(events)), ['a'])
        events = du_backspace+du_a+du_b
        self.assertEqual(list(keyboard.get_typed_strings(events)), ['ab'])
    def test_get_typed_strings_shift(self):
        events = d_shift+du_a+du_b+u_shift+du_space+du_ctrl+du_a
        self.assertEqual(list(keyboard.get_typed_strings(events)), ['AB ', 'a'])
    def test_get_typed_strings_all(self):
        events = du_a+du_b+du_backspace+d_shift+du_a+du_capslock+du_b+u_shift+du_space+du_ctrl+du_a
        self.assertEqual(list(keyboard.get_typed_strings(events)), ['aAb ', 'A'])

    def test_get_hotkey_name_simple(self):
        self.assertEqual(keyboard.get_hotkey_name(['a']), 'a')
    def test_get_hotkey_name_modifiers(self):
        self.assertEqual(keyboard.get_hotkey_name(['a', 'shift', 'ctrl']), 'ctrl+shift+a')
    def test_get_hotkey_name_normalize(self):
        self.assertEqual(keyboard.get_hotkey_name(['SHIFT', 'left ctrl']), 'ctrl+shift')
    def test_get_hotkey_name_plus(self):
        self.assertEqual(keyboard.get_hotkey_name(['+']), 'plus')
    def test_get_hotkey_name_duplicated(self):
        self.assertEqual(keyboard.get_hotkey_name(['+', 'plus']), 'plus')
    def test_get_hotkey_name_full(self):
        self.assertEqual(keyboard.get_hotkey_name(['+', 'left ctrl', 'shift', 'WIN', 'right alt']), 'ctrl+alt+shift+windows+plus')
    def test_get_hotkey_name_multiple(self):
        self.assertEqual(keyboard.get_hotkey_name(['ctrl', 'b', '!', 'a']), 'ctrl+!+a+b')
    def test_get_hotkey_name_from_pressed(self):
        self.do(du_c+d_ctrl+d_a+d_b)
        self.assertEqual(keyboard.get_hotkey_name(), 'ctrl+a+b')

    def test_read_hotkey(self):
        queue = keyboard._queue.Queue()
        def process():
            queue.put(keyboard.read_hotkey())
        from threading import Thread
        t = Thread(target=process)
        t.daemon = True
        t.start()
        time.sleep(0.01)
        self.do(d_ctrl+d_a+d_b+u_ctrl)
        self.assertEqual(queue.get(timeout=0.5), 'ctrl+a+b')

    def test_read_event(self):
        queue = keyboard._queue.Queue()
        def process():
            queue.put(keyboard.read_event(suppress=True))
        from threading import Thread
        t = Thread(target=process)
        t.daemon = True
        t.start()
        time.sleep(0.01)
        self.do(d_a, [])
        self.assertEqual(queue.get(timeout=0.5), d_a[0])

    def test_read_key(self):
        queue = keyboard._queue.Queue()
        def process():
            queue.put(keyboard.read_key(suppress=True))
        from threading import Thread
        t = Thread(target=process)
        t.daemon = True
        t.start()
        time.sleep(0.01)
        self.do(d_a, [])
        self.assertEqual(queue.get(timeout=0.5), 'a')

    def test_wait_infinite(self):
        self.triggered = False
        def process():
            keyboard.wait()
            self.triggered = True
        from threading import Thread
        t = Thread(target=process)
        t.daemon = True # Yep, we are letting this thread loose.
        t.start()
        time.sleep(0.01)
        self.assertFalse(self.triggered)

    def test_wait_until_success(self):
        queue = keyboard._queue.Queue()
        def process():
            queue.put(keyboard.wait(queue.get(timeout=0.5), suppress=True) or True)
        from threading import Thread
        t = Thread(target=process)
        t.daemon = True
        t.start()
        queue.put('a')
        time.sleep(0.01)
        self.do(d_a, [])
        self.assertTrue(queue.get(timeout=0.5))
    def test_wait_until_fail(self):
        def process():
            keyboard.wait('a', suppress=True)
            self.fail()
        from threading import Thread
        t = Thread(target=process)
        t.daemon = True # Yep, we are letting this thread loose.
        t.start()
        time.sleep(0.01)
        self.do(d_b)

    def test_add_hotkey_single_step_suppress_allow(self):
        keyboard.add_hotkey('a', lambda: trigger() or True, suppress=True)
        self.do(d_a, triggered_event+d_a)
    def test_add_hotkey_single_step_suppress_args_allow(self):
        arg = object()
        keyboard.add_hotkey('a', lambda a: self.assertIs(a, arg) or trigger() or True, args=(arg,), suppress=True)
        self.do(d_a, triggered_event+d_a)
    def test_add_hotkey_single_step_suppress_single(self):
        keyboard.add_hotkey('a', trigger, suppress=True)
        self.do(d_a, triggered_event)
    def test_add_hotkey_single_step_suppress_removed(self):
        keyboard.remove_hotkey(keyboard.add_hotkey('a', trigger, suppress=True))
        self.do(d_a, d_a)
    def test_add_hotkey_single_step_suppress_removed(self):
        keyboard.remove_hotkey(keyboard.add_hotkey('ctrl+a', trigger, suppress=True))
        self.do(d_ctrl+d_a, d_ctrl+d_a)
        self.assertEqual(keyboard._listener.filtered_modifiers[dummy_keys['left ctrl'][0][0]], 0)
    def test_remove_hotkey_internal(self):
        remove = keyboard.add_hotkey('shift+a', trigger, suppress=True)
        self.assertTrue(all(keyboard._listener.blocking_hotkeys.values()))
        self.assertTrue(all(keyboard._listener.filtered_modifiers.values()))
        self.assertNotEqual(keyboard._hotkeys, {})
        remove()
        self.assertTrue(not any(keyboard._listener.filtered_modifiers.values()))
        self.assertTrue(not any(keyboard._listener.blocking_hotkeys.values()))
        self.assertEqual(keyboard._hotkeys, {})
    def test_remove_hotkey_internal_multistep_start(self):
        remove = keyboard.add_hotkey('shift+a, b', trigger, suppress=True)
        self.assertTrue(all(keyboard._listener.blocking_hotkeys.values()))
        self.assertTrue(all(keyboard._listener.filtered_modifiers.values()))
        self.assertNotEqual(keyboard._hotkeys, {})
        remove()
        self.assertTrue(not any(keyboard._listener.filtered_modifiers.values()))
        self.assertTrue(not any(keyboard._listener.blocking_hotkeys.values()))
        self.assertEqual(keyboard._hotkeys, {})
    def test_remove_hotkey_internal_multistep_end(self):
        remove = keyboard.add_hotkey('shift+a, b', trigger, suppress=True)
        self.do(d_shift+du_a+u_shift)
        self.assertTrue(any(keyboard._listener.blocking_hotkeys.values()))
        self.assertTrue(not any(keyboard._listener.filtered_modifiers.values()))
        self.assertNotEqual(keyboard._hotkeys, {})
        remove()
        self.assertTrue(not any(keyboard._listener.filtered_modifiers.values()))
        self.assertTrue(not any(keyboard._listener.blocking_hotkeys.values()))
        self.assertEqual(keyboard._hotkeys, {})
    def test_add_hotkey_single_step_suppress_with_modifiers(self):
        keyboard.add_hotkey('ctrl+shift+a', trigger, suppress=True)
        self.do(d_ctrl+d_shift+d_a, triggered_event)
    def test_add_hotkey_single_step_suppress_with_modifiers_fail_unrelated_modifier(self):
        keyboard.add_hotkey('ctrl+shift+a', trigger, suppress=True)
        self.do(d_ctrl+d_shift+u_shift+d_a, d_shift+u_shift+d_ctrl+d_a)
    def test_add_hotkey_single_step_suppress_with_modifiers_fail_unrelated_key(self):
        keyboard.add_hotkey('ctrl+shift+a', trigger, suppress=True)
        self.do(d_ctrl+d_shift+du_b, d_shift+d_ctrl+du_b)
    def test_add_hotkey_single_step_suppress_with_modifiers_unrelated_key(self):
        keyboard.add_hotkey('ctrl+shift+a', trigger, suppress=True)
        self.do(d_ctrl+d_shift+du_b+d_a, d_shift+d_ctrl+du_b+triggered_event)
    def test_add_hotkey_single_step_suppress_with_modifiers_release(self):
        keyboard.add_hotkey('ctrl+shift+a', trigger, suppress=True)
        self.do(d_ctrl+d_shift+du_b+d_a+u_ctrl+u_shift, d_shift+d_ctrl+du_b+triggered_event+u_ctrl+u_shift)
    def test_add_hotkey_single_step_suppress_with_modifiers_out_of_order(self):
        keyboard.add_hotkey('ctrl+shift+a', trigger, suppress=True)
        self.do(d_shift+d_ctrl+d_a, triggered_event)
    def test_add_hotkey_single_step_suppress_with_modifiers_repeated(self):
        keyboard.add_hotkey('ctrl+a', trigger, suppress=True)
        self.do(d_ctrl+du_a+du_b+du_a, triggered_event+d_ctrl+du_b+triggered_event)
    def test_add_hotkey_single_step_suppress_with_modifiers_release(self):
        keyboard.add_hotkey('ctrl+a', trigger, suppress=True, trigger_on_release=True)
        self.do(d_ctrl+du_a+du_b+du_a, triggered_event+d_ctrl+du_b+triggered_event)
    def test_add_hotkey_single_step_suppress_with_modifier_superset_release(self):
        keyboard.add_hotkey('ctrl+a', trigger, suppress=True, trigger_on_release=True)
        self.do(d_ctrl+d_shift+du_a+u_shift+u_ctrl, d_ctrl+d_shift+du_a+u_shift+u_ctrl)
    def test_add_hotkey_single_step_suppress_with_modifier_superset(self):
        keyboard.add_hotkey('ctrl+a', trigger, suppress=True)
        self.do(d_ctrl+d_shift+du_a+u_shift+u_ctrl, d_ctrl+d_shift+du_a+u_shift+u_ctrl)
    def test_add_hotkey_single_step_timeout(self):
        keyboard.add_hotkey('a', trigger, timeout=1, suppress=True)
        self.do(du_a, triggered_event)
    def test_add_hotkey_multi_step_first_timeout(self):
        keyboard.add_hotkey('a, b', trigger, timeout=0.01, suppress=True)
        time.sleep(0.03)
        self.do(du_a+du_b, triggered_event)
    def test_add_hotkey_multi_step_last_timeout(self):
        keyboard.add_hotkey('a, b', trigger, timeout=0.01, suppress=True)
        self.do(du_a, [])
        time.sleep(0.05)
        self.do(du_b, du_a+du_b)
    def test_add_hotkey_multi_step_success_timeout(self):
        keyboard.add_hotkey('a, b', trigger, timeout=0.05, suppress=True)
        self.do(du_a, [])
        time.sleep(0.01)
        self.do(du_b, triggered_event)
    def test_add_hotkey_multi_step_suffix_timeout(self):
        keyboard.add_hotkey('a, b, a', trigger, timeout=0.01, suppress=True)
        self.do(du_a+du_b, [])
        time.sleep(0.05)
        self.do(du_a, du_a+du_b)
        self.do(du_b+du_a, triggered_event)
    def test_add_hotkey_multi_step_allow(self):
        keyboard.add_hotkey('a, b', lambda: trigger() or True, suppress=True)
        self.do(du_a+du_b, triggered_event+du_a+du_b)

    def test_add_hotkey_single_step_nonsuppress(self):
        queue = keyboard._queue.Queue()
        keyboard.add_hotkey('ctrl+shift+a+b', lambda: queue.put(True), suppress=False)
        self.do(d_shift+d_ctrl+d_a+d_b)
        self.assertTrue(queue.get(timeout=0.5))
    def test_add_hotkey_single_step_nonsuppress_repeated(self):
        queue = keyboard._queue.Queue()
        keyboard.add_hotkey('ctrl+shift+a+b', lambda: queue.put(True), suppress=False)
        self.do(d_shift+d_ctrl+d_a+d_b)
        self.do(d_shift+d_ctrl+d_a+d_b)
        self.assertTrue(queue.get(timeout=0.5))
        self.assertTrue(queue.get(timeout=0.5))
    def test_add_hotkey_single_step_nosuppress_with_modifiers_out_of_order(self):
        queue = keyboard._queue.Queue()
        keyboard.add_hotkey('ctrl+shift+a', lambda: queue.put(True), suppress=False)
        self.do(d_shift+d_ctrl+d_a)
        self.assertTrue(queue.get(timeout=0.5))
    def test_add_hotkey_single_step_suppress_regression_1(self):
        keyboard.add_hotkey('a', trigger, suppress=True)
        self.do(d_c+d_a+u_c+u_a, d_c+d_a+u_c+u_a)

    def test_remap_hotkey_single(self):
        keyboard.remap_hotkey('a', 'b')
        self.do(d_a+u_a, d_b+u_b)
    def test_remap_hotkey_complex_dst(self):
        keyboard.remap_hotkey('a', 'ctrl+b, c')
        self.do(d_a+u_a, d_ctrl+du_b+u_ctrl+du_c)
    def test_remap_hotkey_modifiers(self):
        keyboard.remap_hotkey('ctrl+shift+a', 'b')
        self.do(d_ctrl+d_shift+d_a+u_a, du_b)
    def test_remap_hotkey_modifiers_repeat(self):
        keyboard.remap_hotkey('ctrl+shift+a', 'b')
        self.do(d_ctrl+d_shift+du_a+du_a, du_b+du_b)
    def test_remap_hotkey_modifiers_state(self):
        keyboard.remap_hotkey('ctrl+shift+a', 'b')
        self.do(d_ctrl+d_shift+du_c+du_a+du_a, d_shift+d_ctrl+du_c+u_shift+u_ctrl+du_b+d_ctrl+d_shift+u_shift+u_ctrl+du_b+d_ctrl+d_shift)
    def test_remap_hotkey_release_incomplete(self):
        keyboard.remap_hotkey('a', 'b', trigger_on_release=True)
        self.do(d_a, [])
    def test_remap_hotkey_release_complete(self):
        keyboard.remap_hotkey('a', 'b', trigger_on_release=True)
        self.do(du_a, du_b)

    def test_parse_hotkey_combinations_scan_code(self):
        self.assertEqual(keyboard.parse_hotkey_combinations(30), (((30,),),))
    def test_parse_hotkey_combinations_single(self):
        self.assertEqual(keyboard.parse_hotkey_combinations('a'), (((1,),),))
    def test_parse_hotkey_combinations_single_modifier(self):
        self.assertEqual(keyboard.parse_hotkey_combinations('shift+a'), (((1, 5), (1, 6)),))
    def test_parse_hotkey_combinations_single_modifiers(self):
        self.assertEqual(keyboard.parse_hotkey_combinations('shift+ctrl+a'), (((1, 5, 7), (1, 6, 7)),))
    def test_parse_hotkey_combinations_multi(self):
        self.assertEqual(keyboard.parse_hotkey_combinations('a, b'), (((1,),), ((2,),)))
    def test_parse_hotkey_combinations_multi_modifier(self):
        self.assertEqual(keyboard.parse_hotkey_combinations('shift+a, b'), (((1, 5), (1, 6)), ((2,),)))
    def test_parse_hotkey_combinations_list_list(self):
        self.assertEqual(keyboard.parse_hotkey_combinations(keyboard.parse_hotkey_combinations('a, b')), keyboard.parse_hotkey_combinations('a, b'))
    def test_parse_hotkey_combinations_fail_empty(self):
        with self.assertRaises(ValueError):
            keyboard.parse_hotkey_combinations('')


    def test_add_hotkey_multistep_suppress_incomplete(self):
        keyboard.add_hotkey('a, b', trigger, suppress=True)
        self.do(du_a, [])
        self.assertEqual(keyboard._listener.blocking_hotkeys[(1,)], [])
        self.assertEqual(len(keyboard._listener.blocking_hotkeys[(2,)]), 1)
    def test_add_hotkey_multistep_suppress_incomplete(self):
        keyboard.add_hotkey('a, b', trigger, suppress=True)
        self.do(du_a+du_b, triggered_event)
    def test_add_hotkey_multistep_suppress_modifier(self):
        keyboard.add_hotkey('shift+a, b', trigger, suppress=True)
        self.do(d_shift+du_a+u_shift+du_b, triggered_event)
    def test_add_hotkey_multistep_suppress_fail(self):
        keyboard.add_hotkey('a, b', trigger, suppress=True)
        self.do(du_a+du_c, du_a+du_c)
    def test_add_hotkey_multistep_suppress_three_steps(self):
        keyboard.add_hotkey('a, b, c', trigger, suppress=True)
        self.do(du_a+du_b+du_c, triggered_event)
    def test_add_hotkey_multistep_suppress_repeated_prefix(self):
        keyboard.add_hotkey('a, a, c', trigger, suppress=True, trigger_on_release=True)
        self.do(du_a+du_a+du_c, triggered_event)
    def test_add_hotkey_multistep_suppress_repeated_key(self):
        keyboard.add_hotkey('a, b', trigger, suppress=True)
        self.do(du_a+du_a+du_b, du_a+triggered_event)
        self.assertEqual(keyboard._listener.blocking_hotkeys[(2,)], [])
        self.assertEqual(len(keyboard._listener.blocking_hotkeys[(1,)]), 1)
    def test_add_hotkey_multi_step_suppress_regression_1(self):
        keyboard.add_hotkey('a, b', trigger, suppress=True)
        self.do(d_c+d_a+u_c+u_a+du_c, d_c+d_a+u_c+u_a+du_c)
    def test_add_hotkey_multi_step_suppress_replays(self):
        keyboard.add_hotkey('a, b, c', trigger, suppress=True)
        self.do(du_a+du_b+du_a+du_b+du_space, du_a+du_b+du_a+du_b+du_space)

    def test_add_word_listener_success(self):
        queue = keyboard._queue.Queue()
        def free():
            queue.put(1)
        keyboard.add_word_listener('abc', free)
        self.do(du_a+du_b+du_c+du_space)
        self.assertTrue(queue.get(timeout=0.5))
    def test_add_word_listener_no_trigger_fail(self):
        queue = keyboard._queue.Queue()
        def free():
            queue.put(1)
        keyboard.add_word_listener('abc', free)
        self.do(du_a+du_b+du_c)
        with self.assertRaises(keyboard._queue.Empty):
            queue.get(timeout=0.01)
    def test_add_word_listener_timeout_fail(self):
        queue = keyboard._queue.Queue()
        def free():
            queue.put(1)
        keyboard.add_word_listener('abc', free, timeout=1)
        self.do(du_a+du_b+du_c+[make_event(KEY_DOWN, name='space', time=2)])
        with self.assertRaises(keyboard._queue.Empty):
            queue.get(timeout=0.01)
    def test_duplicated_word_listener(self):
        keyboard.add_word_listener('abc', trigger)
        keyboard.add_word_listener('abc', trigger)
    def test_add_word_listener_remove(self):
        queue = keyboard._queue.Queue()
        def free():
            queue.put(1)
        keyboard.add_word_listener('abc', free)
        keyboard.remove_word_listener('abc')
        self.do(du_a+du_b+du_c+du_space)
        with self.assertRaises(keyboard._queue.Empty):
            queue.get(timeout=0.01)
    def test_add_word_listener_suffix_success(self):
        queue = keyboard._queue.Queue()
        def free():
            queue.put(1)
        keyboard.add_word_listener('abc', free, match_suffix=True)
        self.do(du_a+du_a+du_b+du_c+du_space)
        self.assertTrue(queue.get(timeout=0.5))
    def test_add_word_listener_suffix_fail(self):
        queue = keyboard._queue.Queue()
        def free():
            queue.put(1)
        keyboard.add_word_listener('abc', free)
        self.do(du_a+du_a+du_b+du_c)
        with self.assertRaises(keyboard._queue.Empty):
            queue.get(timeout=0.01)

    #def test_add_abbreviation(self):
    #    keyboard.add_abbreviation('abc', 'aaa')
    #    self.do(du_a+du_b+du_c+du_space, [])


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
import struct
from subprocess import check_output
import re
from ._nixcommon import EV_KEY, EV_REL, EV_MSC, EV_SYN, EV_ABS, aggregate_devices
from ._mouse_event import ButtonEvent, WheelEvent, MoveEvent, LEFT, RIGHT, MIDDLE, X, X2, UP, DOWN

import ctypes
import ctypes.util
from ctypes import c_uint32, c_uint, c_int, byref

display = None
window = None
x11 = None
def build_display():
    global display, window, x11
    if display and window and x11: return
    x11 = ctypes.cdll.LoadLibrary(ctypes.util.find_library('X11'))
    # Required because we will have multiple threads calling x11,
    # such as the listener thread and then main using "move_to".
    x11.XInitThreads()
    display = x11.XOpenDisplay(None)
    # Known to cause segfault in Fedora 23 64bits, no known workarounds.
    # http://stackoverflow.com/questions/35137007/get-mouse-position-on-linux-pure-python
    window = x11.XDefaultRootWindow(display)

def get_position():
    build_display()
    root_id, child_id = c_uint32(), c_uint32()
    root_x, root_y, win_x, win_y = c_int(), c_int(), c_int(), c_int()
    mask = c_uint()
    ret = x11.XQueryPointer(display, c_uint32(window), byref(root_id), byref(child_id),
                            byref(root_x), byref(root_y),
                            byref(win_x), byref(win_y), byref(mask))
    return root_x.value, root_y.value

def move_to(x, y):
    build_display()
    x11.XWarpPointer(display, None, window, 0, 0, 0, 0, x, y)
    x11.XFlush(display)

REL_X = 0x00
REL_Y = 0x01
REL_Z = 0x02
REL_HWHEEL = 0x06
REL_WHEEL = 0x08

ABS_X = 0x00
ABS_Y = 0x01

BTN_MOUSE = 0x110
BTN_LEFT = 0x110
BTN_RIGHT = 0x111
BTN_MIDDLE = 0x112
BTN_SIDE = 0x113
BTN_EXTRA = 0x114

button_by_code = {
    BTN_LEFT: LEFT,
    BTN_RIGHT: RIGHT,
    BTN_MIDDLE: MIDDLE,
    BTN_SIDE: X,
    BTN_EXTRA: X2,
}
code_by_button = {button: code for code, button in button_by_code.items()}

device = None
def build_device():
    global device
    if device: return
    device = aggregate_devices('mouse')
init = build_device

def listen(queue):
    build_device()

    while True:
        time, type, code, value, device_id = device.read_event()
        if type == EV_SYN or type == EV_MSC:
            continue

        event = None
        arg = None

        if type == EV_KEY:
            event = ButtonEvent(DOWN if value else UP, button_by_code.get(code, '?'), time)
        elif type == EV_REL:
            value, = struct.unpack('i', struct.pack('I', value))

            if code == REL_WHEEL:
                event = WheelEvent(value, time)
            elif code in (REL_X, REL_Y):
                x, y = get_position()
                event = MoveEvent(x, y, time)

        if event is None:
            # Unknown event type.
            continue

        queue.put(event)

def press(button=LEFT):
    build_device()
    device.write_event(EV_KEY, code_by_button[button], 0x01)

def release(button=LEFT):
    build_device()
    device.write_event(EV_KEY, code_by_button[button], 0x00)

def move_relative(x, y):
    build_device()
    # Note relative events are not in terms of pixels, but millimeters.
    if x < 0:
        x += 2**32
    if y < 0:
        y += 2**32
    device.write_event(EV_REL, REL_X, x)
    device.write_event(EV_REL, REL_Y, y)

def wheel(delta=1):
    build_device()
    if delta < 0:
        delta += 2**32
    device.write_event(EV_REL, REL_WHEEL, delta)


if __name__ == '__main__':
    #listen(print)
    move_to(100, 200)
